
var cameraX = 0, cameraY = 0, cameraZ = 0, cameraEX = 0, cameraEY = 0, cameraEZ = 0,
cameraLength = 3, cameraRot, cameraFov = 2, lastPumpkinYaw = 0, lastPumpkinPitch = 0, lastPumpkinYaw2 = 0, lastPumpkinPitch2 = 0, noLastPumpkinState = true,
keyBindings = {}, upKey,rightKey,downKey,leftKey,
leftClicking = false, rightClicking = false, freeCamera = false, currentTool = 0, flooding = false,
lastToolPoints = new Array(4),
touchPoints = [], touchPointIds = [], lastTouchCX = 0, lastTouchCY = 0, lastTouchRadius = -1,

BACKGROUNDS = ["None","Park","Woods","Snowy","Desert","Hall","City"],
FLOORS = ["None","Planks","Stone","Leaves","Metal"],

TOOL_CARVE = 0, TOOL_SCOOP = 1, TOOL_FILL = 2, TOOL_CLEAR = 3;


function InitializePumpkinCarving() {
	cameraRot = new EulerRotation();
	UpdateCamera();
	
	for (var i = 0; i < lastToolPoints.length; i++) lastToolPoints[i] = 1e8;

	//mouse input events, camera drag + carving
	uiPanel.addEventListener("mousedown",function(e) {
		if (e.target === uiPanel) {
			if (e.button === 2) {
				rightClicking = true;
				e.preventDefault();
				e.stopPropagation();
				
			} else if (e.button === 0) {
				leftClicking = true;
				MouseToPumpkin(e.clientX,e.clientY);
			}
		}
	});
	uiPanel.addEventListener("mouseup",function(e) {
		if (e.target === uiPanel) {
			if (e.button === 2) {
				rightClicking = false;
				e.preventDefault();
				e.stopPropagation();
				
			} else if (e.button === 0) {
				MouseToPumpkin(e.clientX,e.clientY);
				leftClicking = false;
				for (var i = 0; i < 4; i++) lastToolPoints[i] = 1e8;
				if (flooding) StopFlooding();
			}
		}
	});
	uiPanel.addEventListener("mousemove",function(e) {
		if (rightClicking) {//camera rotation drag
			var sc = 3/window.innerHeight;
			cameraRot.Set(cameraRot.roll,cameraRot.pitch-e.movementY*sc,cameraRot.yaw+e.movementX*sc);
			UpdateCamera();
			ClearPathTracer();
			
		} else if (leftClicking) {
			MouseToPumpkin(e.clientX,e.clientY);
		}
	});
	uiPanel.addEventListener("contextmenu",function(e) {
		if (rightClicking || e.target === uiPanel) {//disable right click context menu on canvas
			e.preventDefault();
			e.stopPropagation();
		}
	});
	
	//touch events, camera drag, camera zoom and carving
	uiPanel.addEventListener("touchstart",function(e) {
		var ct = e.changedTouches;
		for (var i = 0; i < ct.length; i++) {
			touchPoints.push(ct[i]);
			touchPointIds.push(ct[i].identifier);
		}
		
		HandleTouchControls();
	});
	uiPanel.addEventListener("touchmove",function(e) {
		var ct = e.changedTouches;
		for (var i = 0; i < ct.length; i++) {
			var cid = touchPointIds.indexOf(ct[i].identifier);
			if (cid !== -1) touchPoints[cid] = ct[i];
		}
		
		HandleTouchControls();
	});
	uiPanel.addEventListener("touchend",function(e) {
		var ct = e.changedTouches;
		for (var i = 0; i < ct.length; i++) {
			var cid = touchPointIds.indexOf(ct[i].identifier);
			if (cid !== -1) {
				touchPoints.splice(cid,1);
				touchPointIds.splice(cid,1);
			}
		}
		if (touchPoints.length < 2) {
			lastTouchRadius = -1;
			for (var i = 0; i < 4; i++) lastToolPoints[i] = 1e8;
			if (flooding) StopFlooding();
		}
	});
	uiPanel.addEventListener("touchcancel",function(e) {
		var ct = e.changedTouches;
		for (var i = 0; i < ct.length; i++) {
			var cid = touchPointIds.indexOf(ct[i].identifier);
			if (cid !== -1) {
				touchPoints.splice(cid,1);
				touchPointIds.splice(cid,1);
			}
		}
		if (touchPoints.length < 2) {
			lastTouchRadius = -1;
			for (var i = 0; i < 4; i++) lastToolPoints[i] = 1e8;
			if (flooding) StopFlooding();
		}
	});
	
	//mouse wheel(camera zoom)
	uiPanel.addEventListener("wheel",function(e) {
		if (e.target === uiPanel && !freeCamera) {
			var ad = Math.sign(e.deltaX)+Math.sign(e.deltaY)+Math.sign(e.deltaZ);
			if (ad) {
				if (ad > 0) cameraLength *= 1.1;
				else cameraLength /= 1.1;
				UpdateCamera();
				ClearPathTracer();
			}
		}
	});
	
	//camera movement key bindings
	upKey = new KeyState(["w","W","ArrowUp"]);
	rightKey = new KeyState(["d","D","ArrowRight"]);
	downKey = new KeyState(["s","S","ArrowDown"]);
	leftKey = new KeyState(["a","A","ArrowLeft"]);
	
	//key events, update key state bindings
	uiPanel.focus();
	uiPanel.addEventListener("keydown",function(e) {
		if (e.target === uiPanel) {
			var kb = keyBindings[e.key];
			if (kb) {
				kb.down = true;
				e.preventDefault();
				e.stopPropagation();
			}
		}
	});
	uiPanel.addEventListener("keyup",function(e) {
		if (e.target === uiPanel) {
			var kb = keyBindings[e.key];
			if (kb) {
				kb.down = false;
				e.preventDefault();
				e.stopPropagation();	
			}
		}
	});
}


//handle key state and bindings
/**@constructor*/
function KeyState(bindings) {
	this.down = false;
	for (var i = 0; i < bindings.length; i++) {
		keyBindings[bindings[i]] = this;
	}
}


function HandleTouchControls() {
	if (touchPoints.length > 1) {//double touch, camera drag + zoom
		var cx = 0, cy = 0, rad = 0;
		for (var i = 0; i < touchPoints.length; i++) {
			var tp = touchPoints[i];
			cx += tp.clientX;
			cy += tp.clientY;
		}
		cx /= touchPoints.length;
		cy /= touchPoints.length;
		
		for (var i = 0; i < touchPoints.length; i++) {
			var tp = touchPoints[i],
				dx = tp.clientX-cx, dy = tp.clientY-cy;
			rad = Math.max(rad,Math.sqrt(dx*dx+dy*dy));
		}
		
		if (lastTouchRadius !== -1) {
			var cdx = cx-lastTouchCX, cdy = cy-lastTouchCY,
				sc = 3/window.innerHeight, ccam = false;;
			if (cdx || cdy) {
				cameraRot.Set(cameraRot.roll,cameraRot.pitch-cdy*sc,cameraRot.yaw+cdx*sc);
				ccam = true;
			}
			
			cdx = lastTouchRadius/Math.max(1e-6,rad);
			if (cdx !== 1 && !freeCamera) cameraLength *= cdx;
			
			if (ccam) {
				UpdateCamera();
				ClearPathTracer();
			}
		}
		
		lastTouchCX = cx;
		lastTouchCY = cy;
		lastTouchRadius = rad;
		
	}
	
	if (touchPoints.length === 1) {//single touch pumpkin draw
		MouseToPumpkin(touchPoints[0].clientX,touchPoints[0].clientY);
	} else {
		for (var i = 0; i < 4; i++) lastToolPoints[i] = 1e8;
		if (flooding) StopFlooding();
	}
}


//Update camera position and rotation matrix.
function UpdateCamera() {
	if (freeCamera) {
		cameraEX = cameraX; cameraEY = cameraY; cameraEZ = cameraZ;
	} else {
		cameraRot.Apply(0,0,-cameraLength);
		cameraEX = fbuf[0]; cameraEY = fbuf[1]; cameraEZ = fbuf[2];
	}
	
	if (floorEnabled) cameraEY = Math.max(cameraEY,0.03-frameParams[31]);//prevent clipping into floor

	frameParams[4] = cameraEX; frameParams[5] = cameraEY; frameParams[6] = cameraEZ;
	
	frameParams[8] = cameraRot.rollCos; frameParams[9] = cameraRot.rollSin;//update camera rotation matrix
	frameParams[10] = cameraRot.pitchCos; frameParams[11] = cameraRot.pitchSin;
	frameParams[12] = cameraRot.yawCos; frameParams[13] = cameraRot.yawSin;
}


/*Class: EulerRotation
Optimized euler angle rotation class, caches cos/sin values.

Constructor: EulerRotation
New EulerRotation.

Parameters:
*float* roll - Roll rotation in radians, if undefined defaults to 0.
*float* pitch - Pitch rotation in radians.
*float* yaw - Yaw rotation in radians.*/
/**@constructor*/
function EulerRotation(roll,pitch,yaw) {
	/*Property: roll
	*float* Roll rotation in radians.*/
	this.roll = roll?roll:0;
	/*Property: pitch
	*float* Pitch rotation in radians.*/
	this.pitch = pitch?pitch:0;
	/*Property: yaw
	*float* Yaw rotation in radians.*/
	this.yaw = yaw?yaw:0;
	
	this.rollCos = Math.cos(this.roll); this.rollSin = Math.sin(this.roll);
	this.pitchCos = Math.cos(this.pitch); this.pitchSin = Math.sin(this.pitch);
	this.yawCos = Math.cos(this.yaw); this.yawSin = Math.sin(this.yaw);
}
/*Property: tmpEuler
<EulerRotation> Temporary euler rotation buffer.*/
var tmpEuler = new EulerRotation();

/*Function: Set
Set euler rotation, you must call this to update cached cos/sin values.

Parameters:
*float* roll - Roll rotation in radians.
*float* pitch - Pitch rotation in radians.
*float* yaw - Yaw rotation in radians.*/
EulerRotation.prototype.Set = function(roll,pitch,yaw) {
	this.roll = roll;
	this.pitch = pitch;
	this.yaw = yaw;
	this.rollCos = Math.cos(roll); this.rollSin = Math.sin(roll);
	this.pitchCos = Math.cos(pitch); this.pitchSin = Math.sin(pitch);
	this.yawCos = Math.cos(yaw); this.yawSin = Math.sin(yaw);
}
/*Function: Apply
Applies euler rotation to vector 'x', 'y', 'z', outputs result in fbuf[0], fbuf[1], fbuf[2].

Parameters:
*float* x - X coordinate.
*float* y - Y coordinate.
*float* z - Z coordinate.*/
EulerRotation.prototype.Apply = function(x,y,z) {
	if (this.roll) {
		fbuf[0] = x;
		x = x*this.rollCos+y*this.rollSin;
		y = y*this.rollCos-fbuf[0]*this.rollSin;
	}
	if (this.pitch) {
		fbuf[1] = y*this.pitchCos+z*this.pitchSin;
		z = z*this.pitchCos-y*this.pitchSin;
	} else fbuf[1] = y;
	if (this.yaw) {
		fbuf[0] = x*this.yawCos+z*this.yawSin;
		fbuf[2] = z*this.yawCos-x*this.yawSin;
	} else {
		fbuf[0] = x;
		fbuf[2] = z;
	}
}
/*Function: ApplyInverse
Applies inverted euler rotation to vector 'x', 'y', 'z', outputs result in fbuf[0], fbuf[1], fbuf[2].

Parameters:
*float* x - X coordinate.
*float* y - Y coordinate.
*float* z - Z coordinate.*/
EulerRotation.prototype.ApplyInverse = function(x,y,z) {
	if (this.yaw) {
		fbuf[0] = x;
		x = x*this.yawCos-z*this.yawSin;
		z = z*this.yawCos+fbuf[0]*this.yawSin;
	}
	if (this.pitch) {
		fbuf[0] = y;
		y = y*this.pitchCos-z*this.pitchSin;
		fbuf[2] = z*this.pitchCos+fbuf[0]*this.pitchSin;
	} else fbuf[2] = z;
	if (this.roll) {
		fbuf[0] = x*this.rollCos-y*this.rollSin;
		fbuf[1] = y*this.rollCos+x*this.rollSin;
	} else {
		fbuf[0] = x;
		fbuf[1] = y;
	}
}


//pumpkin geometry distance function, mirror of geometry function in Pumpkin.glsl
function PumpkinDistance(x,y,z) {
	var pang = Math.atan(x,z)+Math.PI*4+Math.cos(y*15)*0.01+Math.sin(y*44)*0.003,
		plen = Math.sqrt(x*x+z*z), twave = pang*frameParams[20],
		grooves = Math.cos(pang*frameParams[22])*frameParams[23]+Math.cos(pang*frameParams[24])*frameParams[25]+Math.cos(pang*frameParams[26])*frameParams[27]+
				Math.pow(Math.max(Math.abs((twave%1)*2-1)-0.5, 0)*2, 4.0)*frameParams[21];
	y *= frameParams[7];

	return Math.abs(Math.sqrt(x*x+y*y+z*z)-0.95+grooves+Math.pow(Math.min(0.75,Math.abs(y))*Math.max(0,1-plen),2)*frameParams[19])-0.05;
}

//ray march pumpkin, returns true if hit and outputs position in fbuf[0..2]
function PumpkinTrace(px,py,pz, dx,dy,dz) {
	for (var i = 0; i < 100; i++) {
		var dst = PumpkinDistance(px,py,pz);
		if (dst < 0.02) {
			fbuf[0] = px-dx*0.02; fbuf[1] = py-dy*0.02; fbuf[2] = pz-dz*0.02;
			return true; 
		}
		dst *= 0.75;
		px += dx*dst; py += dy*dst; pz += dz*dst;
	}
	return false;
}

//apply carving using mouse movement
function MouseToPumpkin(mx,my) {
	if (flooding) return;

	var dx = (mx/window.innerWidth*2-1)*cameraFov*rtA.aspectX, dy = (my/window.innerHeight*2-1)*cameraFov*-rtA.aspectY, dz = 1,//local ray from screen position
		dlen = 1/Math.sqrt(dx*dx+dy*dy+dz*dz),
		yaw = 1e8, pitch = 1e8;
	dx *= dlen; dy *= dlen; dz *= dlen;
	cameraRot.Apply(dx,dy,dz);//apply camera rotation to ray
	if (PumpkinTrace(cameraEX,cameraEY,cameraEZ,fbuf[0],fbuf[1],fbuf[2])) {//trace ray against pumpkin
		dlen = 1/Math.max(1e-6,Math.sqrt(fbuf[0]*fbuf[0]+fbuf[1]*fbuf[1]+fbuf[2]*fbuf[2]));
		yaw = Math.atan2(fbuf[2],fbuf[0])/(Math.PI*2.)+0.5;//convert hit point to polar coords on pumpkin
		pitch = Math.acos(fbuf[1]*dlen)/Math.PI;
		
		//pass coords and params into tool shader
		if (lastToolPoints[2] >= 1e8) {
			lastToolPoints[2] = yaw; lastToolPoints[3] = pitch;
		}
		for (var i = 0; i < 4; i++) toolParams[i] = lastToolPoints[i];
		toolParams[4] = yaw; toolParams[5] = pitch;
		toolParams[6] = gui["tool_size"].value;
		toolParams[7] = gui["tool_depth"].value;
		if (toolParams[7] > 1) toolParams[7] = 1+Math.pow(toolParams[7]-1,4)*100;
		
		if (currentTool > 1) toolParams[6] = currentTool===TOOL_FILL?1:0;
		
		pumpkinHeightB.Bind();//execute tool shader
		var tsh = toolShaders[Math.min(2,currentTool)];
		tsh.Use();
		gl.uniform4fv(tsh["params"],toolParams);
		BindTexture(tsh["map"],pumpkinHeightA.texture);
		gl.drawArrays(gl.TRIANGLES,0,3);
		UnbindAllTextures();
		var tmp = pumpkinHeightA;
		pumpkinHeightA = pumpkinHeightB;
		pumpkinHeightB = tmp;
		
		if (currentTool > 1) flooding = true;
		ClearPathTracer();
	}
	lastToolPoints[0] = lastToolPoints[2]; lastToolPoints[1] = lastToolPoints[3];
	lastToolPoints[2] = yaw; lastToolPoints[3] = pitch;//buffer of last mouse movements
}

//clear pumpkin map of flood state
function StopFlooding() {
	pumpkinHeightB.Bind();
	floodResetShader.Use();
	BindTexture(floodResetShader["map"],pumpkinHeightA.texture);
	gl.drawArrays(gl.TRIANGLES,0,3);
	UnbindAllTextures();
	var tmp = pumpkinHeightA;
	pumpkinHeightA = pumpkinHeightB;
	pumpkinHeightB = tmp;
		
	flooding = false;
}


//update loop
function Update() {
	//free camera movement
	if (freeCamera) {
		var moveX = 0, moveZ = 0;
		if (upKey.down) moveZ++;
		if (downKey.down) moveZ--;
		if (rightKey.down) moveX++;
		if (leftKey.down) moveX--;
	
		if (moveX || moveZ) ClearPathTracer();
	
		cameraRot.Apply(moveX*3,0,moveZ*3);
		cameraX += fbuf[0]*delta;
		cameraY += fbuf[1]*delta;
		cameraZ += fbuf[2]*delta;
	
		UpdateCamera();
	}
	
	//flood filling/clearing
	if (flooding) {
		pumpkinHeightB.Bind();//execute flood tool shader
		var tsh = toolShaders[2];
		tsh.Use();
		gl.uniform4fv(tsh["params"],toolParams);
		BindTexture(tsh["map"],pumpkinHeightA.texture);
		gl.drawArrays(gl.TRIANGLES,0,3);
		UnbindAllTextures();
		var tmp = pumpkinHeightA;
		pumpkinHeightA = pumpkinHeightB;
		pumpkinHeightB = tmp;
		
		if (frameParams[2] > 10/1000) ClearPathTracer();
	}
}
