
var rtA, rtB, pumpkinShaders, pumpkinFloorShader, displayShader, displayTransShader, copyShader, toolShaders = new Array(3), floodResetShader,
frameParams = new Float32Array(10*4), toolParams = new Float32Array(8),
pumpkinHeightA, pumpkinHeightB, pumpkinHeightLast, background, floorCol, floorHeight, floorRough,
floorEnabled = false, capturingRender = false, renderNumSamples = 0, renderSamples = 0, backgroundId = 0, pixelScaling = 1;


function InitializeRendering() {
	//setup webgls
	if (!InitializeQuickWebGL(document.getElementsByTagName("canvas")[0])) {
		FatalCrash("Fatal Error: Your system does not support WebGL 2.");
		return false;
	}
	if (!gl.getExtension("OES_texture_float")) {
		FatalCrash("Fatal Error: Your system does not support floating point render textures.");
		return false;
	}

	//initial opengl settings
	gl.enable(gl.CULL_FACE);
	gl.cullFace(gl.BACK);
	gl.disable(gl.DEPTH_TEST);
	gl.depthMask(false);
	gl.clearDepth(1.0);
	gl.disable(gl.STENCIL_TEST);
	gl.disable(gl.BLEND);
	gl.enable(gl.SCISSOR_TEST);
	gl.colorMask(true,true,true,true);
	
	//allocate resources
	displayShader = GetShader("Display.glsl");
	displayTransShader = GetShader("DisplayTransparent.glsl");
	pumpkinShaders = GetShader("Pumpkin.glsl","FLOOR_ENABLED","TRANSPARENT_ENABLED");
	copyShader = GetShader("Copy.glsl");
	toolShaders[0] = GetShader("Knife.glsl");
	toolShaders[1] = GetShader("Scoop.glsl");
	toolShaders[2] = GetShader("FloodFillClear.glsl");
	floodResetShader = GetShader("FloodReset.glsl");
	
	rtA = new RenderTexture(1,1);
	rtB = new RenderTexture(1,1);
	pumpkinHeightA = new RenderTexture(1024,512);
	SetTextureOptions(true,false,true,true);
	pumpkinHeightB = new RenderTexture(1024,512);
	SetTextureOptions(true,false,true,true);
	pumpkinHeightLast = new RenderTexture(1024,512);
	
	background = new ImageTexture(true,false,true,true);
	floorCol = new ImageTexture(true,false,true,true);
	floorHeight = new ImageTexture(true,false,true,true,true);
	floorRough = new ImageTexture(true,false,true,true,true);
	
	//clear default pumpkin mask
	pumpkinHeightA.Bind();
	gl.clearColor(1,0,0,1);
	gl.clear(gl.COLOR_BUFFER_BIT);

	//window size event listener
	OnWindowResize();
	window.addEventListener("resize",OnWindowResize);
	
	return true;
}

//texture loaded from image element
/**@constructor*/
function ImageTexture(lin,mip,rx,ry,grayscale) {
	this.image = new Image();
	this.image.imageTextureObj = this;
	this.image.addEventListener("load",OnImageTextureLoad);
	this.texture = gl.createTexture();
	this.grayscale = grayscale?true:false;
	gl.bindTexture(gl.TEXTURE_2D, this.texture);
	SetTextureOptions(lin,mip,rx,ry);
}
function OnImageTextureLoad(e) {
	var obj = e.target.imageTextureObj, form = obj.grayscale?gl.LUMINANCE:gl.RGBA,
		wd = obj.image.width, ht = obj.image.height,
		nwd = 1, nht = 1,
		timg = obj.image;
	//resize non power of 2 image to prevent issues
	while (nwd < wd) nwd *= 2;
	while (nht < ht) nht *= 2;
	if (nwd !== wd || nht !== ht) {
		var tcanv = document.createElement("canvas"),
			t2d = tcanv.getContext("2d");
		tcanv.width = nwd; tcanv.height = nht;
		t2d.drawImage(timg,0,0,nwd,nht);
		timg = tcanv;
	}
	
	gl.bindTexture(gl.TEXTURE_2D,obj.texture);
	gl.texImage2D(gl.TEXTURE_2D,0,form,form,gl.UNSIGNED_BYTE,timg);
	ClearPathTracer();
}

//create new shader with text content from element extra arguments are #define options that results in multiple shader versions being returned
function GetShader(ele) {
	var src = document.getElementById(ele).textContent;
	if (arguments.length > 1) {
		var ocount = arguments.length-1,
			scount = 1<<ocount,
			sa = new Array(scount);
		for (var i = 0; i < scount; i++) {
			var nsrc = src;
			for (var o = 0; o < ocount; o++) {
				if (((i>>o)&1) === 0) {
					var astr = "#define "+arguments[o+1];
					nsrc = nsrc.replace(astr,"//"+astr);
				}
			}
			sa[i] = new Shader(nsrc);
		}
		return sa;
	} else {
		return new Shader(src);
	}
}

function ClearPathTracer() {
	if (!capturingRender) frameParams[2] = 0;
}

function SetFloor(id) {
	if (id) {
		floorEnabled = true;
		var pt = "Floors/"+FLOORS[id];
		floorCol.image.src = pt+"_Color.jpg";
		floorHeight.image.src = pt+"_Height.jpg";
		floorRough.image.src = pt+"_Rough.jpg";
	} else {
		floorEnabled = false;
		ClearPathTracer();
	}
}
function SetBackground(id) {
	if (id) {
		background.image.src = "Backgrounds/"+BACKGROUNDS[id]+".jpg";
	}
	var last = backgroundId;
	backgroundId = id;
	if ((last===0) !== (id===0)) UpdateBackgroundParams();
}

function UpdateBackgroundParams() {
	var exposure = backgroundId===0 ? 0 : Math.pow(bgcfg["bg_exposure"].value,2),
		col = bgcfg["bg_color"].value;
	frameParams[16] = Math.pow(col.r/255,2.2)*exposure;
	frameParams[17] = Math.pow(col.g/255,2.2)*exposure;
	frameParams[18] = Math.pow(col.b/255,2.2)*exposure;
	ClearPathTracer();
}

function ComputeFloorOffset() {
	frameParams[31] = 0.75+(1/frameParams[7]-0.8)*0.8//pumpkin height
	-frameParams[14]*0.35//floor height
	-frameParams[19]*0.1;//indents
}

function ComputePumpkinBounds() {
	var bv = Math.max(1.05/frameParams[7]-frameParams[19],1.2);
	frameParams[35] = bv*bv;
}

function ComputeLightColor() {
	var col = bgcfg["light_color"].value, str = bgcfg["light_strength"].value;
	frameParams[36] = Math.pow(col.r/255,2.2)*str;
	frameParams[37] = Math.pow(col.g/255,2.2)*str;
	frameParams[38] = Math.pow(col.b/255,2.2)*str;
}


//on window resize callback
function OnWindowResize() {
	if (!capturingRender) {
		//handle dpi scaling
		var ndpi = window.devicePixelRatio;
		if (!ndpi) ndpi = 1;
	
		//compute canvas size
		quickGLCanvas.width = display.width = Math.floor(window.innerWidth*ndpi);
		quickGLCanvas.height = display.height = Math.floor(window.innerHeight*ndpi);
		if (ndpi !== 1 || pixelScaling !== 1) {
			quickGLCanvas.style.width = window.innerWidth+"px";
			quickGLCanvas.style.height = window.innerHeight+"px";
		}
		pixelScaling = ndpi;
	
		//apply to render textures
		rtA.SetSize(display.width,display.height,gl.RGBA,gl.RGBA,gl.FLOAT);
		rtB.SetSize(display.width,display.height,gl.RGBA,gl.RGBA,gl.FLOAT);
		frameParams[0] = cameraFov*rtA.aspectX;
		frameParams[1] = cameraFov*rtA.aspectY;
		frameParams[3] = (2*cameraFov)/(display.width+display.height);
	
		ClearPathTracer();
	}
}



//render frame
function Render() {
	//path trace
	rtA.Bind();
	var rtransparent, pts;
	if (capturingRender) rtransparent = rendercfg["render_trans"].value;
	pts = pumpkinShaders[(floorEnabled?1:0) + (rtransparent?2:0)];
	pts.Use();
	gl.uniform4fv(pts["params"], frameParams);
	BindTexture(pts["buffer"], rtB.texture);
	BindTexture(pts["background"], background.texture);
	BindTexture(pts["pumpkin"], pumpkinHeightA.texture);
	if (floorEnabled) {
		BindTexture(pts["floorCol"], floorCol.texture);
		BindTexture(pts["floorHeight"], floorHeight.texture);
		BindTexture(pts["floorRough"], floorRough.texture);
	}
	gl.drawArrays(gl.TRIANGLES,0,3);
	UnbindAllTextures();
	
	if (capturingRender) {
		//update render progress
		if (renderSamples%20 === 0) {
			renderState.textContent = "Rendered "+renderSamples+"/"+renderNumSamples+" samples...";
		}
		renderSamples++;
		if (renderSamples >= renderNumSamples) {
			//complete
			TEV(renderingPanel);
			TEV(resultPanel);
			
			//render to canvas so we can grab data url
			quickGLCanvas.width = display.width = rtA.width;
			quickGLCanvas.height = display.height = rtA.height;
			display.Bind();
			var ds = rtransparent?displayTransShader:displayShader;
			ds.Use();
			if (rtransparent) gl.uniform1f(ds["samples"],renderNumSamples);
			BindTexture(ds["buffer"], rtA.texture);
			gl.drawArrays(gl.TRIANGLES,0,3);
			UnbindAllTextures();
			
			//get and set base64 image result
			var rimg = document.getElementById("result_img");
			rimg.src = quickGLCanvas.toDataURL();
			
			capturingRender = false;
			OnWindowResize();
		}
				
	} else {
		//display
		display.Bind();
		displayShader.Use();
		BindTexture(displayShader["buffer"], rtA.texture);
		gl.drawArrays(gl.TRIANGLES,0,3);
		UnbindAllTextures();
	}
	
	//swap buffers
	var tmp = rtA;
	rtA = rtB;
	rtB = tmp;
	frameParams[2] += 1/1000;
}

/*
measure fps via console

Update();
var frames = 1000,
	startTime = Date.now();
for (var i = 0; i < frames; i++) Render();
gl.finish();
var endTime = Date.now();
console.log(frames/((endTime-startTime)/1000)+" fps");


floor - 36.94 fps
floor w/ early break - 36.74 fps
early break(dynamic branching) doesnt have benefits in webgl 1 :(
*/
